package xCAT_plugin::alt;

use Storable qw(dclone);
use Sys::Syslog;
use File::Temp qw/tempdir/;
use xCAT::Table;
use xCAT::Utils;
use xCAT::SvrUtils;
use xCAT::MsgUtils;
use Data::Dumper;
use xCAT::NetworkUtils;
use Getopt::Long;
Getopt::Long::Configure("bundling");
Getopt::Long::Configure("pass_through");
use File::Path;
use File::Copy;
use File::Basename;
#use strict;
my @cpiopid;

sub handled_commands
{
    return {
            copycd    => "alt",
            mknetboot => "nodetype:os=(alt.*)",
            mkinstall => "nodetype:os=(alt.*)",
            mkstatelite => "nodetype:os=(alt.*)",
            };
}

sub process_request
{
    my $request  = shift;
    my $callback = shift;
    my $doreq    = shift;
    my $distname = undef;
    my $arch     = undef;
    my $path     = undef;
    if ($request->{command}->[0] eq 'copycd')
    {
        return copycd($request, $callback, $doreq);
    }
    elsif ($request->{command}->[0] eq 'mkinstall')
    {
        return mkinstall($request, $callback, $doreq);
    }
    elsif ($request->{command}->[0] eq 'mknetboot' or
	$request->{command}->[0] eq 'mkstatelite')
    {
        return mknetboot($request, $callback, $doreq);
    }
}

sub copycd
{
    my $request  = shift;
    my $callback = shift;
    my $doreq    = shift;
    my $distname = "";
    my $detdistname = "";
    my $installroot;
    my $arch;
    my $path;

    # info($callback, "copycd", "called");

    # Cleanup at unexpexted situations
    $SIG{INT} =  $SIG{TERM} = sub {
       foreach(@cpiopid){
          kill 2, $_; 
       }
       if ($::CDMOUNTPATH) {
            chdir("/");
            system("umount $::CDMOUNTPATH");
       }
    };

    $installroot = xCAT::Utils->getInstallDir();

    @ARGV = @{$request->{arg}};
    GetOptions(
               'n=s' => \$distname,
               'a=s' => \$arch,
               'p=s' => \$path
               );
    unless ($path)
    {
        # We need to know where to get ISO contents
        return;
    }
    if ($distname and $distname !~ /alt/)
    {
        # This module attended to work with ALT Linux distros
        return;
    }

    my $repoinfo = $path . "/ALTLinux/base/release.main";

    unless (-r $repoinfo)
    {
        # Can't fetch information about ALT Linux CD
        return;
    }

    # Disk info
    my $dinfo;

    # Open disk info file
    open($dinfo, $repoinfo);

    my $dver;
    my $darch;

    # Parse repository info file
    while (<$dinfo>)
    {
        # Fetch target architecture
        if (m/^Architecture:\s+(\S+)/i)
        {
            $darch = $1;
            chomp($darch);
            last;
        }

        # Fetch ALT branch version
        if (m/^Version:\s+(\S+)/i)
        {
            $dver = $1;
            chomp($dver);
            last;
        }
    }
    close($dinfo);

    # For broken CD information, failback to system arch
    unless ($darch) {
        $darch = `uname -m`;
        chomp ($darch);
    }

    # Translate from CPU family to architecture
    if ($darch and $darch =~ /i.86/)
    {
        $darch = "x86";
    }

    # Suppose x86_64 for any unknown architectures
    unless ($darch)
    {
        # Fail back to x86_64 architecture
        $darch = "x86_64"
    }

    # Expected architecture by user must confirm to existing
    if ($darch)
    {
        unless ($arch)
        {
            $arch = $darch;
        }
        if ($arch and $arch ne $darch)
        {
            $callback->(
                     {
                      error =>
                        ["Requested architecture $arch, but media is $darch"],
                        errorcode => [1]
                     }
                     );
            return;
        }
    }

    # Branch version is mandatory to know
    unless ($dver)
    {
            $callback->({error => "Unknow ALT Linux branch at media"});
            return;
    }

    $distname = "alt";

    # Check for previously grabbed disks
    if (-d "$installroot/$distname/$arch") {
        $callback->({error => "Found earlier copied ALT Linux distro for: $arch"});
        return;
    }

    # Clear request we have got it
    %{$request} = ();

    # Lets user know what it is going on
    $callback->({data => "Copying ALT Linux media to $installroot/$distname/$arch/"});


    # Set up umask for further file operations
    my $omask = umask 0022;
    umask $omask;

    mkpath("$installroot/$distname/$arch");
    chmod 0755, "$installroot/$distname/$arch";


    my $rc;
    my $kid;

    # Go where source media mounted
    chdir $path;

    # Total files at media
    my $numFiles = `find . -print | wc -l`;


    # Go multiprocess and communicate between yourselves
    # Forks a child connected to the have opened filehandle
    my $child = open($kid,"|-");

    unless (defined $child) {
      $callback->({error=>"Media copy operation fork failure"});
      return;
    }

    if ($child) {
        # Parent - say files to copy
        push @cpiopid,$child;
        my @finddata = `find .`;
        for (@finddata) {
            # Print filename to child
            print $kid $_;
        }
        close($kid);
        $rc = $?;
    } else {
        # Child - listen filenames to copy from STDIN
        my $c = "nice -n 20 cpio -vdump $installroot/$distname/$arch";
        # Cpio pipes output to us
        my $k2 = open(PIPE, "$c 2>&1 |") ||
            $callback->({error => "Media copy operation fork failure"});
        push @cpiopid, $k2;
        my $copied = 0;
        my ($percent, $fout);
        # Calculate current percent using information from cpio
        while(<PIPE>){
            next if /^cpio:/;
            $percent = $copied / $numFiles;
            $fout = sprintf "%0.2f%%", $percent * 100;
            $callback->({sinfo => "$fout"});
            ++$copied;
        }
        exit;
    }


    if ($rc != 0)
    {
        $callback->({error => "Media copy operation failed, status $rc"});
    }
    else
    {
        $callback->({data => "Media copy operation successful"});
        my @ret=xCAT::SvrUtils->update_tables_with_templates($distname, $arch);
        if ($ret[0] != 0) {
            $callback->({data => "Error when updating the osimage tables: " . $ret[1]});
        }

    }
}

sub info {
    my $callback = shift;
    my $stage    = shift;
    my $msg      = shift;

    if (1) {
        $callback->({ warning => ["($stage) $msg"]});
    }
}

sub mkinstall
{
    require xCAT::Template;

    my $platform = "alt";
    my $request  = shift;
    my $callback = shift;
    my $doreq    = shift;
    my @nodes    = @{$request->{node}};
    my %img_hash=();

    # Used tables
    my $ostab         = xCAT::Table->new('nodetype');
    my $restab        = xCAT::Table->new('noderes');
    my $bptab         = xCAT::Table->new('bootparams',-create=>1);
    my $hmtab         = xCAT::Table->new('nodehm');
    my $osimagetab    = xCAT::Table->new('osimage', -create=>1);
    my $linuximagetab = xCAT::Table->new('linuximage', -create=>1);

    my $installroot = xCAT::Utils->getInstallDir();
    my $tftpdir = xCAT::Utils->getTftpDir();

    my $node;

    # Node attributes
    # `nodetype' table
    my %osents = %{$ostab->getNodesAttribs(\@nodes, ['profile', 'os', 'arch', 'provmethod'])};
    # `noderes' table
    my %rents = %{$restab->getNodesAttribs(\@nodes, ['nfsserver', 'primarynic', 'installnic'])};
    # `nodehm' table
    my %hents = %{$hmtab->getNodesAttribs(\@nodes, ['serialport', 'serialspeed', 'serialflow'])};


    # Parse each node
    foreach $node (@nodes)
    {
        my $os;
        my $arch;
        my $profile;
        my $tmplfile;
        my $pkglistfile;
        my $restpkgs;
        my $pkgdir;
        my $imagename;
        my $osinst;

        my $ent = $osents{$node}->[0];

        # nodetype.provmethod - provisioning method for node deployment.
        # Possible values:
        #  * install
        #  * netboot
        #  * OS image from the osimage table

        # nodetype.provmethod == OS image
        if ($ent
                and $ent->{provmethod}
                and ($ent->{provmethod} ne 'install')
                and ($ent->{provmethod} ne 'netboot')) {

            # OS image from the osimage table
            $imagename=$ent->{provmethod};
            info($callback, "mkinstall", "imagename = $imagename");

            if (!exists($img_hash{$imagename})) {
                (my $ref) = $osimagetab->getAttribs({imagename => $imagename}, 'osvers', 'osarch', 'profile', 'provmethod');
                if ($ref) {
                    # Fetch values from `osimage' table for node
                    $img_hash{$imagename}->{osver}      = $ref->{'osvers'};
                    $img_hash{$imagename}->{osarch}     = $ref->{'osarch'};
                    $img_hash{$imagename}->{profile}    = $ref->{'profile'};
                    $img_hash{$imagename}->{provmethod} = $ref->{'provmethod'};

                    # Fetch values from `linuximage' table
                    (my $ref1) = $linuximagetab->getAttribs({imagename => $imagename}, 'template', 'pkgdir', 'pkglist');
                    if ($ref1) {
                        if ($ref1->{'template'}) {
                            $img_hash{$imagename}->{template} = $ref1->{'template'};
                        }
                        if ($ref1->{'pkgdir'}) {
                            $img_hash{$imagename}->{pkgdir}   = $ref1->{'pkgdir'};
                        }
                        if ($ref1->{'pkglist'}) {
                            $img_hash{$imagename}->{pkglist}=$ref1->{'pkglist'};
                        }
                    }
                    # if the install template wasn't found, then lets look for it in the default locations.
                    unless ($img_hash{$imagename}->{template}) {
                        my $tmplfile = xCAT::SvrUtils::get_tmpl_file_name("$installroot/custom/install/alt",
                            $ref->{'profile'}, $ref->{'osvers'}, $ref->{'osarch'}, $ref->{'osvers'});
                        if (! $tmplfile) {
                            $tmplfile=xCAT::SvrUtils::get_tmpl_file_name("$::XCATROOT/share/xcat/install/alt",
                                $ref->{'profile'}, $ref->{'osvers'}, $ref->{'osarch'}, $ref->{'osvers'});
                        }
                        # if we managed to find it, put it in the hash:
                        if($tmplfile){
                            $img_hash{$imagename}->{template}=$tmplfile;
                        }
                    }
                } else {
                    $callback->(
                        {error     => ["The os image $imagename does not exists on the osimage table for $node"],
                            errorcode => [1]});
                    next;
                }
            }

            my $ph = $img_hash{$imagename};

            # Values from `osimage' table
            $os       = $ph->{osver};   # Example: alt
            $arch     = $ph->{osarch};  # Example: x86_64
            $profile  = $ph->{profile}; # Example: compute

            # Values from `linuximage' table
            $tmplfile   = $ph->{template};  # Example: /usr/share/xcat/install/alt/compute.tmpl
            $pkgdir     = $ph->{pkgdir};    # Example: /install/alt/x86_64
            $pkglistfile= $ph->{pkglist};

            if (!$pkgdir) {
                $pkgdir = "$installroot/$os/$arch";
            }
        }
        # nodetype.provmethod == install or netboot
        else {
            # Values from `osimage' table
            $os      = $ent->{os};
            $arch    = $ent->{arch};
            $profile = $ent->{profile};

            # First lookup for customized template
            $tmplfile = xCAT::SvrUtils::get_tmpl_file_name("$installroot/custom/install/$platform", $profile, $os, $arch);

            # Lookup for generic template
            if (! $tmplfile) {
                $tmplfile = xCAT::SvrUtils::get_tmpl_file_name("$::XCATROOT/share/xcat/install/$platform", $profile, $os, $arch);
                info($callback, "mkinstall", "Found template $tmplfile for $platform, $os, $arch");
            }

            # First lookup for customized packages list
            $pkglistfile=xCAT::SvrUtils::get_pkglist_file_name("$installroot/custom/install/$platform", $profile, $os, $arch);

            # Lookup for generic package list
            unless ($pkglistfile ) {
                $pkglistfile=xCAT::SvrUtils::get_pkglist_file_name("$::XCATROOT/share/xcat/install/$platform", $profile, $os, $arch);
            }

            info($callback, "mkinstall", "profile     = $profile");
            info($callback, "mkinstall", "os          = $os");
            info($callback, "mkinstall", "arch        = $arch");
            info($callback, "mkinstall", "tmplfile    = $tmplfile");
            if ($pkglistfile) {
                info($callback, "mkinstall", "pkglistfile = $pkglistfile");
            } else {
                info($callback, "mkinstall", "pkglistfile = EMPTY. Use standard packages list from CD/DVD disk.");
            }

            $pkgdir="$installroot/$os/$arch";
        }

        # Check for presents all necessary parameters
        my @missingparms;
        unless ($os) {
            if ($imagename) { push @missingparms,"osimage.osvers";  }
            else { push @missingparms,"nodetype.os";}
        }
        unless ($arch) {
            if ($imagename) { push @missingparms,"osimage.osarch";  }
            else { push @missingparms,"nodetype.arch";}
        }
        unless ($profile) {
            if ($imagename) { push @missingparms,"osimage.profile";  }
            else { push @missingparms,"nodetype.profile";}
        }
        unless ($os and $arch and $profile)
        {
            $callback->(
                {
                    error => ["Missing ".join(',',@missingparms)." for $node"],
                    errorcode => [1]
                }
            );
            next;    #No profile
        }

        my $tmperr;

        # Is template file exists ?
        unless ( -r "$tmplfile")
        {
            if ($imagename) {
                $tmperr="Unable to find template file: $tmplfile";
            } else {
                $tmperr="No $platform kickstart template exists for "
                . $profile
                . " in $installroot/custom/install/$platform or $::XCATROOT/share/xcat/install/$platform (for $profile/$os/$arch combination)";
            }

            $callback->({
                         error => [ "$tmperr" ],
                         errorcode => [1]
                        });
            next;
        }

        # Call the Template class to do substitution to produce a kickstart file in the autoinst dir
        mkpath("$installroot/autoinst/$node");
        $tmperr = xCAT::Template->subvars($tmplfile, "$installroot/autoinst/$node/autoinstall.scm", $node);
        if ($tmperr)
        {
            $callback->({
                         node => [{name => [$node], error => [$tmperr], errorcode => [1]}]
                        });
            next;
        }

        my $tmpldir = dirname("$tmplfile");

        # Copy pre intallation script
        my $preinst = "$tmpldir/../scripts/pre.alt";
        if ( -f "$preinst" ) {
            $tmperr = xCAT::Template->subvars($preinst, "$installroot/autoinst/$node/xcat-pre.sh", $node);
            if ($tmperr)
            {
                $callback->({
                        node => [{name => [$node], error => [$tmperr], errorcode => [1]}]
                    });
                next;
            }
        }

        # Copy post intallation script
        my $postinst = "$tmpldir/../scripts/post.alt";
        if ( -f "$postinst" ) {
            $tmperr = xCAT::Template->subvars($postinst, "$installroot/autoinst/$node/xcat-post.sh", $node);
            if ($tmperr)
            {
                $callback->({  
                        node => [{name => [$node], error => [$tmperr], errorcode => [1]}]
                    });
                next;
            }
        }

        # Copy template for alterator-vm
        unlink ("$installroot/autoinst/$node/vm-profile.scm");
        if ( -f "$tmpldir/vm-profile.scm" ) {
            copy("$tmpldir/vm-profile.scm", "$installroot/autoinst/$node/vm-profile.scm");
        }

        # Additional packages list defined by administrator.
        # WARNING: some packages may be missed at ISO image. Use for your own risk.
        # *.installer.pkglist lists is unofficial xCAT future. It is ALT Linux specific.
        unless ($restpkgs) {
            $restpkgs = my_get_profile_def_filename($os, $profile, $arch, "$installroot/custom/install/alt", "installer.pkglist");
            unless ($restpkgs) {
                $restpkgs = my_get_profile_def_filename($os, $profile, $arch, "$::XCATROOT/share/xcat/install/alt", "installer.pkglist");
            }
        }
        $pkggrp_src = "$pkgdir/Metadata/pkg-groups.tar";
        $pkggrp_dst = "$installroot/autoinst/$node/pkg-groups.tar";

        unless ( -f "$pkggrp_src" ) {
            info($callback, "mkinstall", "Unable to find generic package list $pkggrp_src for $profile.");
        }

        # Remove previous Metadata
        if ( -f "$pkggrp_dst" ) {
            unlink ($pkggrp_dst);
        }

        if ( -f $restpkgs ) {
            info($callback, "mkinstall", "Additional packages listed at $restpkgs, will be installed during node provision");

            # Copy standard packages lists and groups to node configuration directory
            copy("$pkggrp_src", "$pkggrp_dst");

            # Add additional packages list
            #    . q{ --transform 's,^.*/\([^/]\+\)$,lists/\1,'}
            system("tar"
                . " --append"
                . " --file $pkggrp_dst"
                . q{ --transform 's,^.*$,./lists/xcat,'}
                . " $restpkgs");
        } else {
            info($callback, "mkinstall", "Unable to find additional package list for `$profile' profile. It is ok.");
        }

        # Copy ssh root key, later this key will be used in ALT Linux installer
        my $pubkey = "$installroot/postscripts/_ssh/authorized_keys";
        # Remove previous installed key, if any
        unlink ("$installroot/autoinst/$node/root.pub");
        if (-f "$pubkey") {
            copy ("$pubkey", "$installroot/autoinst/$node/root.pub");
        }

        # ALT Linux kernel location and first stage
        my $kernpath   = "$pkgdir/syslinux/alt0/vmlinuz";
        my $initrdpath = "$pkgdir/syslinux/alt0/full.cz";

        if ((! -r "$kernpath") or (! -r "$initrdpath"))
        {
            $callback->(
                {
                    error => ["Install image not found in $pkgdir/syslinux/alt0"],
                    errorcode => [1]
                }
            );
        }

        my %doneimgs;

        unless ($doneimgs{"$os|$arch"})
        {
            mkpath("$tftpdir/xcat/$os/$arch");
            copy($kernpath,"$tftpdir/xcat/$os/$arch");
            copy($initrdpath,"$tftpdir/xcat/$os/$arch");
            $doneimgs{"$os|$arch"} = 1;
        }

        # `noderes' table values
        my $rent  = $rents{$node}->[0];

        # `nodehm' table values
        my $sent = $hents{$node}->[0];

        unless ($rent and $rent->{nfsserver})
        {
            $callback->(
                {
                    error => ["No noderes.nfsserver defined for " . $node],
                    errorcode => [1]
                }
            );
            next;
        }

        my $nic;

        if ($rent->{installnic})
        {
            if ($rent->{installnic} eq "mac")
            {
                my $mactab = xCAT::Table->new("mac");
                my $macref = $mactab->getNodeAttribs($node, ['mac']);
                $nic = $macref->{mac};
            }
            else
            {
                $nic = $rent->{installnic};
            }
        }
        elsif ($rent->{primarynic})
        {
            if ($rent->{primarynic} eq "mac")
            {
                my $mactab = xCAT::Table->new("mac");
                my $macref = $mactab->getNodeAttribs($node, ['mac']);
                $nic = $macref->{mac};
            }
            else
            {
                $nic = $rent->{primarynic};
            }
        }
        else
        {
            $nic = "$eth0";
        }

        # Why it is blank, no mac defined?
        if ($nic eq "")
        {
            $callback->(
                {
                    error => ["No MAC address defined for " . $node],
                    errorcode => [1]
                }
            );
        }

        my $kcmdline = "automatic=method:http"
                       . ",interface:$nic"
                       . ",network:dhcp"
                       . ",server:$rent->{nfsserver}"
                       . ",directory:/install/alt/$arch"
                       . " ai"
                       . " xcat" # flag to run xCAT postscripts
                       . " curl=http://$rent->{nfsserver}/install/autoinst/$node/"
                       . " ramdisk_size=165536";

# XXX@stanv . "/install/autoinst/" . $node;

        if (defined($sent->{serialport}))
        {
            unless ($sent->{serialspeed})
            {
                $callback->(
                    {
                        error => [
                        "serialport defined, but no serialspeed for $node in nodehm table"
                        ],
                        errorcode => [1]
                    }
                );
                next;
            }
            $kcmdline .=
            " console=tty0 console=ttyS"
            . $sent->{serialport} . ","
            . $sent->{serialspeed};
            if ($sent->{serialflow} =~ /(hard|cts|ctsrts)/)
            {
                $kcmdline .= "n8r";
            }
        }
        $kcmdline .= " noipv6";
        $bptab->setNodeAttribs(
            $node,
            {
                kernel   => "xcat/$os/$arch/vmlinuz",
                initrd   => "xcat/$os/$arch/full.cz",
                kcmdline => $kcmdline
            }
        );
    }
}

sub mknetboot
{
    my $req      = shift;
    my $callback = shift;
    my $doreq    = shift;
    my $nodes    = @{$req->{node}};
    my @nodes    = @{$req->{node}};
    my $ostab    = xCAT::Table->new('nodetype');
    my $linuximagetab;
    my $osimagetab;
    my $statelite; # inverse to $stateless
    my $stateless; # inverse to $statelite
    my $statetab;
    my $stateHash;

    my $installroot = xCAT::Utils->getInstallDir();
    my $tftpdir = xCAT::Utils->getTftpDir();
    my $sitetab  = xCAT::Table->new('site');

    my $xcatdport = "3001";

    if ($sitetab)
    {
        ($ref) = $sitetab->getAttribs({key => 'xcatdport'}, 'value');
        if ($ref and $ref->{value})
        {
            $xcatdport = $ref->{value};
        }
    }

    if($req->{command}->[0] =~ 'mkstatelite'){
        info($callback, "mknetboot", "Generate statelite image\n");
        $statelite = 1;
        $stateless = not $statelite;
    } else {
        info($callback, "mknetboot", "Generate stateless image\n");
        $statelite = 0;
        $stateless = not $statelite;
    }

    if ($statelite) {
        $statetab = xCAT::Table->new('statelite', -create=>1);
        $stateHash = $statetab->getNodesAttribs(\@nodes, ['statemnt']);
    }

    my $mactab = xCAT::Table->new('mac');
    my $machash = $mactab->getNodesAttribs(\@nodes, ['interface', 'mac']);

    my $restab = xCAT::Table->new('noderes');
    my $reshash = $restab->getNodesAttribs(\@nodes, ['primarynic', 'tftpserver', 'xcatmaster', 'nfsserver',' nfsdir', 'installnic']);

    my $ntents = $ostab->getNodesAttribs($req->{node}, ['os', 'arch', 'profile', 'provmethod']);
    my %img_hash=();
    my %donetftp=();

    foreach my $node (@nodes)
    {
        my $osver;
        my $arch;
        my $profile;
        my $rootimgdir;
        my $nodebootif; # nodebootif will be used if noderes.installnic is not set

        my $ent= $ntents->{$node}->[0];

        if ($ent and $ent->{provmethod} and ($ent->{provmethod} ne 'install') and ($ent->{provmethod} ne 'netboot') and ($ent->{provmethod} ne 'statelite')) {
            my $imagename=$ent->{provmethod};
            if (!exists($img_hash{$imagename})) {
                if (!$osimagetab) {
                    $osimagetab=xCAT::Table->new('osimage', -create=>1);
                }
                (my $ref) = $osimagetab->getAttribs({imagename => $imagename}, 'osvers', 'osarch', 'profile', 'provmethod');
                if ($ref) {
                    $img_hash{$imagename}->{osver}=$ref->{'osvers'};
                    $img_hash{$imagename}->{osarch}=$ref->{'osarch'};
                    $img_hash{$imagename}->{profile}=$ref->{'profile'};
                    $img_hash{$imagename}->{provmethod}=$ref->{'provmethod'};
                    if (!$linuximagetab) {
                        $linuximagetab=xCAT::Table->new('linuximage', -create=>1);
                    }
                    (my $ref1) = $linuximagetab->getAttribs({imagename => $imagename}, 'rootimgdir', 'nodebootif');
                    if (($ref1) && ($ref1->{'rootimgdir'})) {
                        $img_hash{$imagename}->{rootimgdir}=$ref1->{'rootimgdir'};
                    }
                    if (($ref1) && ($ref1->{'nodebootif'})) {
                        $img_hash{$imagename}->{nodebootif} = $ref1->{'nodebootif'};
                    }
                } else {
                    $callback->(
                        {error     => ["The os image $imagename does not exists on the osimage table for $node"],
                            errorcode => [1]});
                    next;
                }
            }
            my $ph=$img_hash{$imagename};
            $osver = $ph->{osver};
            $arch  = $ph->{osarch};
            $profile = $ph->{profile};

            $nodebootif = $ph->{nodebootif};

            $rootimgdir=$ph->{rootimgdir};
            if (!$rootimgdir) {
                $rootimgdir="$installroot/netboot/$osver/$arch/$profile";
            }
        }
        else {
            $osver = $ent->{os};
            $arch    = $ent->{arch};
            $profile = $ent->{profile};
            $rootimgdir="$installroot/netboot/$osver/$arch/$profile";
        }

        unless ($osver and $arch and $profile)
        {
            $callback->(
                {
                    error     => ["Insufficient nodetype entry or osimage entry for $node"],
                    errorcode => [1]
                }
            );
            next;
        }

        info($callback, "mknetboot", "osvr=$osver, arch=$arch, profile=$profile, imgdir=$rootimgdir\n");

        my $platform="alt";
        my $suffix  = 'gz';

        if (-r "$rootimgdir/rootimg.sfs")
        {
            $suffix = 'sfs';
        }

        # statelite images are not packed

        unless ( -r "$rootimgdir/kernel") {
            $callback->({
                    error=>[qq{Did you run "genimage" before running "packimage"? $rootimgdir/kernel cannot be found for node $node}],
                    errorcode=>[1]
                });
            next;
        }

        # Statelite mode
        if ($statelite) {
            unless (-r "$rootimgdir/initrd-statelite.gz") {
                unless ( -r "$rootimgdir/initrd.gz") {
                    $callback->({
                            error=>[qq{Did you run "genimage" before running "liteimg"? initrd.gz or initrd-statelite.gz cannot be found}],
                            errorcode=>[1]
                        });
                    next;
                }
                info($callback, "mknetboot", "node = $node, copy $rootimgdir/initrd.gz to $rootimgdir/initrd-statelite.gz");
                copy("$rootimgdir/initrd.gz", "$rootimgdir/initrd-statelite.gz");
            }
        }

        # Stateless mode
        if ($stateless) {
            unless (-r "$rootimgdir/initrd-stateless.gz") {
                unless (-r "$rootimgdir/initrd.gz") {
                    $callback->({
                            error=>[qq{Did you run "genimage" before running "packimage"? initrd.gz or initrd-stateless.gz cannot be found}],
                            errorcode=>[1]
                        });
                    next;
                }
                info($callback, "mknetboot", "node = $node, copy $rootimgdir/initrd.gz to $rootimgdir/initrd-stateless.gz");
                copy("$rootimgdir/initrd.gz", "$rootimgdir/initrd-stateless.gz");
            }

            unless ( -r "$rootimgdir/rootimg.gz" or -r "$rootimgdir/rootimg.sfs" ) {
                $callback->({
                        error=>[qq{No packed image for platform $osver, architecture $arch, and profile $profile, please run packimage before nodeset}],
                        errorcode=>[1]
                    });
                next;
            }
        }

        mkpath("/$tftpdir/xcat/netboot/$osver/$arch/$profile/");

        unless ($donetftp{$osver,$arch,$profile}) {
            copy("$rootimgdir/kernel", "/$tftpdir/xcat/netboot/$osver/$arch/$profile/");
            info($callback, "mknetboot", "node = $node, copy $rootimgdir/kernel to /$tftpdir/xcat/netboot/$osver/$arch/$profile/");

            if ($statelite) {
                copy("$rootimgdir/initrd-statelite.gz", "/$tftpdir/xcat/netboot/$osver/$arch/$profile/");
                info($callback, "mknetboot", "node = $node, copy $rootimgdir/initrd-statelite.gz to /$tftpdir/xcat/netboot/$osver/$arch/$profile/");
            }
            if ($stateless) {
                copy("$rootimgdir/initrd-stateless.gz", "/$tftpdir/xcat/netboot/$osver/$arch/$profile/");
                info($callback, "mknetboot", "node = $node, copy $rootimgdir/initrd-stateless.gz to /$tftpdir/xcat/netboot/$osver/$arch/$profile/");
            }
            $donetftp{$osver,$arch,$profile} = 1;
        }

        # Check for kernel image:
        unless ( -r "/$tftpdir/xcat/netboot/$osver/$arch/$profile/kernel" ) {
            $callback->({
                    error=>[qq{Can't find /$tftpdir/xcat/netboot/$osver/$arch/$profile/kernel}],
                    errorcode=>[1]
                });
            next;
        }

        # Check for statelite initrd:
        if ($statelite and ! -r "/$tftpdir/xcat/netboot/$osver/$arch/$profile/initrd-statelite.gz") {
            $callback->({
                    error=>[qq{Can't find /$tftpdir/xcat/netboot/$osver/$arch/$profile/initrd-statelite.gz}],
                    errorcode=>[1]
                });
            next;
        }

        # Check for stateless initrd:
        if ($stateless and ! -r "/$tftpdir/xcat/netboot/$osver/$arch/$profile/initrd-stateless.gz") {
            $callback->({
                    error=>[qq{Can't find /$tftpdir/xcat/netboot/$osver/$arch/$profile/initrd-stateless.gz}],
                    errorcode=>[1]
                });
            next;
        }

        my $bptab  = xCAT::Table->new('bootparams',-create=>1);
        my $hmtab  = xCAT::Table->new('nodehm');
        my $sent   = $hmtab->getNodeAttribs($node, ['serialport', 'serialspeed', 'serialflow']);
        my $ient   = $reshash->{$node}->[0];

        unless ($ient) {
            info($callback, "mknetboot", "Can't fetch information from `noderes' table for node = $node.");
            next;
        }

        # determine image server, if tftpserver use it, else use xcatmaster
        # last resort use self
        my $imgsrv;
        my $xcatmaster;


        if ($ient->{xcatmaster})
        {
            $xcatmaster = $ient->{xcatmaster};
        } else {
            $xcatmaster = '!myipfn!'; #allow service nodes to dynamically nominate themselves as a good contact point, this is of limited use in the event that xcat is not the dhcp/tftp server
        }

        if ($ient->{tftpserver})
        {
            $imgsrv = $ient->{tftpserver};
        }
        else
        {
            $imgsrv = $xcatmaster;
        }

        unless ($imgsrv)
        {
            $callback->(
                {
                    error => [
                    "Unable to determine or reasonably guess the image server for $node"
                    ],
                    errorcode => [1]
                }
            );
            next;
        }

        my $kcmdline = "fastboot ";
        if($statelite){
            # get entry for nfs root if it exists:
            # have to get nfssvr and nfsdir from noderes table
            my $nfssrv = $imgsrv;
            my $nfsdir = $rootimgdir;

            if ($ient->{nfsserver}) {
                $nfssrv = $ient->{nfsserver};
            }
            info($callback, "mknetboot", "node = $node, imgsrv = $nfssrv, nfssrv = $nfssrv");

            if ($ient->{nfsdir} ne '') {
                $nfsdir = $ient->{nfsdir} . "/netboot/$osver/$arch/$profile";
                if($ient->{nfsdir} =~ m!^//!) {
                    $nfsdir = $ient->{nfsdir};
                    $nfsdir =~ s!^/!!;
                }
            }
            info($callback, "mknetboot", "node = $node, rootimgdir = $rootimgdir, nfsdir = $nfsdir");

            $kcmdline .= "NFSROOT=$nfssrv:$nfsdir STATEMNT=";

            # Lookup to `statelite.statemnt' for $node
            # The persistant read/write area where a node's persistent files will be written to, e.g: 10.0.0.1/state/
            my $statemnt = "";
            if (exists($stateHash->{$node})) {
                $statemnt = $stateHash->{$node}->[0]->{statemnt};
                if (grep /\$/, $statemnt) {
                    my ($server, $dir) = split(/:/, $statemnt);

                    #if server is blank, then its the directory
                    unless($dir) {
                        $dir = $server;
                        $server = '';
                    }
                    if(grep /\$|#CMD/, $dir) {
                        $dir = xCAT::SvrUtils->subVars($dir, $node, 'dir', $callback);
                        $dir = ~ s/\/\//\//g;
                    }
                    if($server) {
                        $server = xCAT::SvrUtils->subVars($server, $node, 'server', $callback);
                    }
                    $statemnt = $server . ":" . $dir;
                }
            }
            $kcmdline .= $statemnt . " ";

            info($callback, "mknetboot", "node = $node, persistant read/write area = $statemnt");


            # BEGIN service node
            my $isSV = xCAT::Utils->isServiceNode();
            my $res = xCAT::Utils->runcmd("hostname", 0);
            my $sip = xCAT::NetworkUtils->getipaddr($res);  # this is the IP of service node
            if($isSV and (($xcatmaster eq $sip) or ($xcatmaster eq $res))) {
                # if the NFS directory in litetree is on the service node,
                # and it is not exported, then it will be mounted automatically
                xCAT::SvrUtils->setupNFSTree($node, $sip, $callback);
                # then, export the statemnt directory if it is on the service node
                if($statemnt) {
                    xCAT::SvrUtils->setupStatemnt($sip, $statemnt, $callback);
                }
            }
            # END service node
        }
        else
        {
            $kcmdline .=
            "imgurl=http://$imgsrv/install/netboot/$osver/$arch/$profile/rootimg.$suffix ";
        }

        # get "xcatmaster" value from the "noderes" table
        $kcmdline .= "XCAT=$xcatmaster:$xcatdport ";

        # add the kernel-booting parameter: netdev=<eth0>, or BOOTIF=<mac>
        # netdev= - /proc/cmdline for initrd stateless / statelite
        # BOOTIF= - /proc/cmdline for initrd stateless / statelite
        my $netdev = "";
        my $mac = $machash->{$node}->[0]->{mac};

        # TODO: currently, only "mac" attribute with classic style is used, the "|" delimited string of "macaddress!hostname" format is not used
        if ( (index($mac, "|") eq -1) and (index($mac, "!") eq -1) ) {
            #convert to linux format
            if ($mac !~ /:/) {
                $mac =~s/(..)(..)(..)(..)(..)(..)/$1:$2:$3:$4:$5:$6/;
            }
            $kcmdline .= "$mac ";
        } else {
            info($callback, "mknetboot", qq{In the "mac" table, the "|" delimited string of "macaddress!hostname" format is not supported if installnic/primarynic is set".});
            # Don't pass mac to /proc/cmdline for this $node
        }

        if ($ient->{installnic}) {
            if ($ient->{installnic} ne "mac") {
                $kcmdline .= "netdev=" . $ient->{installnic} . " ";
                info($callback, "mknetboot", "node = $node, (noderes.installnic) netdev=" . $ient->{installnic});
            }
        } elsif ($nodebootif) {
            $kcmdline .=  "netdev=" . $nodebootif . " ";
            info($callback, "mknetboot", "node = $node, (linuximage.nodebootif) netdev=$nodebootif");
        } elsif ($ient->{primarynic}) {
            $kcmdline .= "netdev=" . $ient->{primarynic} . " ";
            info($callback, "mknetboot", "node = $node, (noderes.primarynic) netdev=" . $ient->{primarynic});
        } else {
            if ($arch =~ /x86/) {
                #do nothing, we'll let pxe/xnba work their magic
            } elsif ($mac) {
                $kcmdline .=  "BOOTIF=" . $mac . " ";
                info($callback, "mknetboot", "node = $node, (mac.mac) BOOTIF=$mac");
            } else {
                $callback->({
                        error=>[qq{"cannot get the mac address for $node in mac table"}],
                        errorcode=>[1]
                    });
            }
        }

        if (defined $sent->{serialport})
        {

            #my $sent = $hmtab->getNodeAttribs($node,['serialspeed','serialflow']);
            unless ($sent->{serialspeed})
            {
                $callback->(
                    {
                        error => [
                        "serialport defined, but no serialspeed for $node in nodehm table"
                        ],
                        errorcode => [1]
                    }
                );
                next;
            }
            $kcmdline .=
            "console=tty0 console=ttyS" . $sent->{serialport} . "," . $sent->{serialspeed};
            if ($sent->{serialflow} =~ /(hard|tcs|ctsrts)/)
            {
                $kcmdline .= "n8r";
            }
        }

        # Initrd image name
        my $initrdstr;

        if ($stateless) {
            $initrdstr = "xcat/netboot/$osver/$arch/$profile/initrd-stateless.gz";
        }
        if ($statelite) {
            $initrdstr = "xcat/netboot/$osver/$arch/$profile/initrd-statelite.gz";
        }

        $bptab->setNodeAttribs(
            $node,
            {
                kernel => "xcat/netboot/$osver/$arch/$profile/kernel",
                initrd => $initrdstr,
                kcmdline => $kcmdline
            }
        );
    } # END foreach $node
}

sub my_get_profile_def_filename {
    my $osver = shift;
    my $profile = shift;
    my $arch = shift;

    my $base=shift;
    my $ext=shift;
    my $dotpos = rindex($osver, ".");
    my $osbase = substr($osver, 0, $dotpos);
    if (-r "$base/$profile.$osver.$arch.$ext") {
        return "$base/$profile.$osver.$arch.$ext";
    } elsif (-r "$base/$profile.$osbase.$arch.$ext") {
        return "$base/$profile.$osbase.$arch.$ext";
    } elsif (-r "$base/$profile.$arch.$ext") {
        return "$base/$profile.$arch.$ext";
    } elsif (-r "$base/$profile.$osver.$ext") {
        return "$base/$profile.$osver.$ext";
    } elsif (-r "$base/$profile.$osbase.$ext") {
        return "$base/$profile.$osbase.$ext";
    } elsif (-r "$base/$profile.$ext") {
        return "$base/$profile.$ext";
    }

    return "";
}


1;
